import datetime
import fitz
import logging
import os
import sys

from lost_cat.parsers.base_parser import BaseParser

logger = logging.getLogger(__name__)

class PDFParser(BaseParser):
    """Process a word document"""
    def __init__(self, uri: str = None, bytes_io: bytes = None, settings: dict = None) -> None:
        super().__init__(uri=uri, bytes_io=bytes_io, settings=settings)
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = PDFParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Settings: %s", self.settings)

        # file
        self._uri = None
        self._file = None
        if uri:
            self._uri = uri
            self._file = fitz.open(self._uri)
        elif bytes_io:
            self._file = fitz.open(stream=bytes_io, filetype="pdf")

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{},
            "uritypes": ["file"],
            "source":[
                {
                    "table": "URIMD",
                    "key": "ext",
                    "values": [".pdf"]
                }
            ]
        }

    def parser(self) -> dict:
        """will parser the open file and retrn the result"""
        _data = {}
        for _mdk, _mdv in self.get_metadata().items():
            _mdk = self._alias_tags.get(_mdk, _mdk)
            _data[_mdk] = _mdv
        return _data

    def get_toc(self) -> dict:
        """ """
        _data = []
        _data = self._file.get_toc(simple=True)

        return {
            "type": "toc",
            "toc": _data
        }

    def get_content(self) -> dict:
        """ """
        _pages = {}
        _lines = []
        for _idx, _page in enumerate(self._file):
            _text = _page.get_text('text', sort=True, flags=2)
            if _text:
                _pages[_idx] = _text
                # the text is split on lines
                _lines.extend([x.strip() for x in _text.split("\n") if x.strip()])

        return {
            "type": "lines",
            "pages": _pages,
            "lines": _lines
        }

    def get_metadata(self) -> dict:
        """ """
        _data = {}
        for _mdk, _mdv in self._file.metadata.items():
            if _mdk in ["creationDate", "modDate"]:
                _data[_mdk] = datetime.datetime.strptime(_mdv.replace("'", ""), "D:%Y%m%d%H%M%S%z").strftime("%Y-%m-%d %H:%M:%S")
            else:
                _data[_mdk] = _mdv

        return _data

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "toc": self.get_toc,
            "content": self.get_content,
            "metadata": self.get_metadata,
            "parser": self.parser,
        }

    def close(self, force: bool = False, block: bool = False, timeout: int = -1):
        """will close the """
        if self._file:
            self._file = None



import tarfile
import zipfile
import logging
import os
import sys

from lost_cat.parsers.base_parser import BaseParser
from lost_cat.utils.path_utils import func_switch_zip, get_file_metadata

logger = logging.getLogger(__name__)

class ZipParser(BaseParser):
    """Process a ZIP document"""
    def __init__(self, uri: str = None, bytes_io: bytes = None, settings: dict = None) -> None:
        super().__init__(uri=uri, bytes_io=bytes_io, settings=settings)
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = ZipParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Settings: %s", self.settings)

        # file
        self._uri = None
        self._file = None
        if uri:
            self._uri = uri
            logger.info("File: %s", self._uri)
            self._filemd = get_file_metadata(uri=self._uri, options=self.settings.get("options",{}))
            logger.info("MD: %s", self._filemd)
            self._file = func_switch_zip(self._filemd.get("ext"), "open")(uri = self._uri)
            op_func = func_switch_zip(self._filemd.get("ext"), "scan")
            if op_func:
                self._files = op_func(uri=self._uri)
            else:
                raise NotImplementedError()
        elif bytes_io:
            raise NotImplementedError()

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "toc": self.get_toc,
            "metadata": self.get_metadata,
            "parser": self.parser,
            "content": self.get_content,
            "get_item": self.get_item,
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{
                "splitfolders": True,
                "splitextension": True,
                "stats": True,
            },
            "uritypes": ["file"],
            "source":[
                {
                    "table": "URIMD",
                    "key": "ext",
                    "values": [".zip"]
                }
            ]
        }

    def get_toc(self):
        """ """
        return [d["path"] for d in self._files]

    def get_metadata(self):
        """ """
        _data = {}
        for _kv, _vv in self._filemd.items():
            _data[_kv] = _vv

    def get_content(self):
        """ """
        _data = {}
        for _f in self._files:
            _data[_f.get("path")] = _f
        return _data

    def get_item(self, path:str) -> bytes:
        """ """

        return func_switch_zip(self._filemd.get("ext"), "fetch")(file_obj=self._file, item_path=path)

    def close(self, force: bool = False, block: bool = False, timeout: int = -1):
        """will close the """
        if self._file:
            self._file = None

    def parser(self) -> dict:
        """will parser the open file and retrn the result"""
        _data = {
            "files": self._files
        }
        for _kv, _vv in self._filemd.items():
            _data[_kv] = _vv

        return _data


"""this will scan a word set and return the details"""
from email.errors import ObsoleteHeaderDefect
import io
import logging
import multiprocessing as mp
import os
import threading as td
import zipfile
from queue import Empty

from lost_cat.processors.base_processor import BaseProcessor
from lost_cat.parsers.word_parser import WordParser
from lost_cat.utils.path_utils import SourceNotHandled, scan_files
from lost_cat.utils.tag_anon import TagAnon

logger = logging.getLogger(__name__)

class WordProcessor(BaseProcessor):
    """Will process the queue and provide a """

    def __init__(self, settings: dict = None):
        """"""
        super().__init__(settings=settings)
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"
        self._semiphore = f"DONE: {self._name}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = WordParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Semiphore: %s", self._semiphore)
        logger.debug("Settings: %s", self.settings)

        # init the tags lists
        self._anonobj = None
        self._groups_tags = {}
        self._metadata_tags = {}
        self._alias_tags = {}

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "parser": self.parser,
            #"anonimizer": self.anonimizer,
            #"tags_alias": self.alias_tags,
            #"tags_groups": self.groups_tags,
            #"tags_metadata": self.metadata_tags,
            #"metadata": self.metadata,
            #"contents": self.get_image,
            "export": {
                "numpy": {
                    "ext": ".pickle",
                    #"func": self.get_numpy},
                },
                "hounsfield": {
                    "ext": ".bmp",
                    #"func": self.get_hounsfield,
                }
            }
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{
            },
            "uritypes": [
                "file", "zipfiles"
            ],
            "source":[
                {
                    "table": "URIMD",
                    "field": "key",
                    "select": "ext",
                    "filter": [".docx"],

                }
            ],
            "uri_metadata": ["zipfile"],
            "threads": {
                "count": 1,
                "stop": True, # tells the system to stop when the in queue is empty
                "timeout": 2
            }
        }

    def add_action(self, obj: object = None
            ):
        """ Will add an action to run against the file in the queue
        abd add the return to the metadata if needed."""
        pass

    def anonimizer(self, anonimizer: TagAnon) -> None:
        """Allows for the metadata tags to be anonymized"""
        self._anonobj = anonimizer

    def alias_tags(self, tags: list) -> None:
        """Allows for the metadata tags to be anonymized"""
        self._alias_tags = tags

    def groups_tags(self, tags: list) -> None:
        """Sets the metadata tags to be used for grouping
        The grouping is used to organize the structure"""
        self._groups_tags = tags

    def metadata_tags(self, tags: list):
        """Sets the tags to use for general metadata"""
        self._metadata_tags = tags

    def parser(self) -> None:
        """ load the for each match to a parser, load the
        an in_queue with values, then call the *_processer
        to process the queue.
        The parser enabled processer will use a helper file to handle the files"""
        use_threads = self.settings.get("threads",{}).get("count",5)

        for t_idx in range(use_threads):
            logger.info("Thread: %s",t_idx)
            scan_q = td.Thread(target=self.parser_file)
            scan_q.start()
            scan_q.join()

    def parser_file(self):
        """A parser scanner function, quick and dirty"""
        t_settings = self.settings.get("threads",{})
        self.input.put(self.semiphore)

        while self.input:
            try:
                q_item = self.input.get(timeout=t_settings.get("timeout")) if self.input else None
                if q_item == self.semiphore:
                    break

                # if the user wants to kill the queue if there are not entries...
                # requires timeout set, otherwise the queue get blocks...
                if not q_item and self.settings.get("threads",{}).get("stop"):
                    break

                _uri=q_item.get("uri")
                if not os.path.exists(_uri):
                    # most likely a zipfile...
                    zf = zipfile.ZipFile(q_item.get("zipfile"))
                    file_data = zf.read(q_item.get("uri"))
                    bytes_io = io.BytesIO(file_data)
                    _fileobj = WordParser(bytes_io=bytes_io)

                else:
                    _fileobj = WordParser(uri=_uri)

                # set the anonimizer and tags if defined
                if  _fn := _fileobj.avail_functions().get("anonimizer"):
                    _fn(anonimizer=self._anonobj)

                #for _tn in ["alias", "groups", "metadata"]:
                if _fn := _fileobj.avail_functions().get("alias"):
                    _fn(tags=self._alias_tags)
                if _fn := _fileobj.avail_functions().get("groups"):
                    _fn(tags=self._groups_tags)
                if _fn := _fileobj.avail_functions().get("metadata"):
                    _fn(tags=self._metadata_tags)

                _filemd = _fileobj.parser()
                _fileobj.close()

                # create the returning object
                _data = {
                    "processorid": self.processorid,
                    "uri id": q_item.get("uriid"),
                    "uri_type": q_item.get("uri_type"),
                    "uri": q_item.get("uri"),
                    "metadata": {},
                    "versions": {
                        "__latest__": True,
                        "versionmd" : _filemd
                    }
                }

                logger.debug("O Data: %s", _data)
                self.output.put(_data)

            except Empty:
                break


"""A module to mange the parseing of sets of lines
    into a table, based in field detection """
import logging
import nltk
import re

logger = logging.getLogger(__name__)

nltk.download([
        "names",
        "stopwords",
        "averaged_perceptron_tagger",
        "vader_lexicon",
        "punkt",
    ])

def clean_text(text: str, operations: dict = None) -> list:
    """This will clean a text block and return
    a list of phrases"""
    _data = []

    if operations:
        for _opk, _opv in operations.items():
            if _opv.get("stage","") == "pre":
                if _reg := _opv.get("regex"):
                    text = re.sub(pattern=_reg, repl=_opv.get("replace",""), string=text, flags=re.IGNORECASE)
                else:
                    _fnd = _opv.get("find", "")
                    text = text.replace(_fnd, _opv.get("replace",""))

    for _tt in nltk.sent_tokenize(text):
        if operations:
            for _opk, _opv in operations.items():
                if _opv.get("stage","") == "sentence":
                    if _reg := _opv.get("regex"):
                        _tt = re.sub(pattern=_reg, repl=_opv.get("replace",""), string=_tt, flags=re.IGNORECASE)
                    else:
                        _fnd = _opv.get("find", "")
                        _tt = _tt.replace(_fnd, _opv.get("replace",""))

        _data.append(_tt.strip())
    return _data

def tablify(lines: list, fields: dict) -> dict:
    """ will take the collection of key words / phrases
    and return the text as a key value doctionary
        fields:
            "fieldname" :{
                match:      [<array of string to match>]
                regex:      [<array of regex to get groups]
                groups:     [list of group name in regex]
                fstring:    format string to match on,
                            use "<fstring>".format(**<groups>)
                extract:    add this line as a field, but keep the current
                            field in play
            }
    """
    _data = {}
    _compiled = {}

    _fld_name = None
    _fld_lines = []

    for _l in lines:
        _continue = False
        for _fn, _fd in fields.items():
            for _mt in _fd.get("match", []):
                if _mt == _l.lower():
                    if _fld_name:
                        _data[_fld_name] = "\n".join(_fld_lines)
                    _fld_name = _fn
                    _fld_lines = []
                    _continue = True
                    break

            if _continue:
                logger.debug("String Match: %s", _l)
                break

            for _idx, _restr in enumerate(_fd.get("regex", [])):
                if f"{_fn}:{_idx:3}" in _compiled:
                    _rg = _compiled.get(f"{_fn}:{_idx:3}")
                else:
                    _rg = re.compile(_restr, re.IGNORECASE)
                    _compiled[f"{_fn}:{_idx:3}"] = _rg

                if _m := _rg.match(_l):
                    if _fd.get("extract"):
                        _grps = {}
                        for _g in _fd.get("groups",[]):
                            _grps[_g] = _m.group(_g)
                        if _grps:
                            _data[_fn] = _fd.get("fstring","").format(**_grps)
                        else:
                            _data[_fn] = _l
                        _continue = True
                        break

                    if _fld_name:
                        _data[_fld_name] = "\n".join(_fld_lines)
                    _continue = True
                    _fld_name = _fn
                    _fld_lines = []

                    _grps = {}
                    for _g in _fd.get("groups",[]):
                        _grps[_g] = _m.group(_g)
                    if _grps:
                        _fld_lines.append(_fd.get("fstring","").format(**_grps))

            if _continue:
                logger.debug("Regex Match: %s", _l)
                break

        if _fld_name and not _continue:
            logger.debug("field: %s\tline: %s", _fld_name, _l)
            _fld_lines.append(_l)

    if _fld_name:
        logger.debug("closing: %s", _fld_name)
        _data[_fld_name] = "\n".join(_fld_lines)

    return _data



""""""
import os
import logging
import PIL.Image, PIL.ExifTags
from lost_cat.parsers.base_parser import BaseParser

from lost_cat.utils.tag_anon import TagAnon

logger = logging.getLogger(__name__)

class EXIFParser(BaseParser):
    """

    ---
    Attributes
    ----------
    Methods
    -------

    """

    def __init__(self, uri: str = None, bytes_io: bytes = None, settings: dict = None) -> None:
        super().__init__(uri=uri, bytes_io=bytes_io, settings=settings)
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"

        if not settings:
            logger.debug("Loading default settings")
            #self.settings = DICOMParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Settings: %s", self.settings)

        # file
        self._uri = None
        self._file = None
        if uri:
            self._uri = uri
            #self._file = pydicom.dcmread(self._uri)
        elif bytes_io:
            #self._file = pydicom.dcmread(bytes_io)

    def __str__(self):
        return f"{self.name} {self.version} <{self._uri}>"

    def close(self) -> None:
        """"""
        if self._img:
            self._img = None

    def get_extensions(self):
        """

        Parameters
        ----------
        Returns
        -------
        None
        """
        return [".bmp",".dds",".dib",".eps",".gif",".icns",".ico",".im",".jpeg",".msp",".pcx",".png",".ppm",
        ".sgi",".tga",".tiff",".webp",".xbm",".blp",".cur",".dcx",".fli",".flc",".fpx",".ftex",".gbr",".gd",
        ".imt",",iptc",".naa",".mcidas",".mic",".mpo",".pcd",".pixar",".psd",".wal",".wmf",",xpm"]

    def get_functions(self):
        """

        Parameters
        ----------
        Returns
        -------
        None
        """
        return {
            "metadata": self.get_metadata,
            "analyze": self.analyze,
            #"contents": self.get_contents
        }

    def get_metadata(self) -> dict:
        """This will return the doc info infomation from the
        Named file.

        Parameters
        ----------
        Returns
        -------
        None
        """
        width, height = self._img.size
        tags = {PIL.ExifTags.TAGS[k]: v for k, v in self._img._getexif().items() if k in PIL.ExifTags.TAGS}
        data = {
            "width": width,
            "height": height,
            "tags": tags,
            "mode": self._img.mode,
            "info": self._img.info,
            "bands": list(self._img.getbands())
        }
        return data

    def analyze(self) -> dict:
        """

        Parameters
        ----------
        Returns
        -------
        None
        """
        data = dict

        return data

    def get_contents(self) -> list:
        """This will return the paragroah objects in a word document

        Parameters
        ----------
        Returns
        -------
        None
        """
        data = []

        return data


"""this will scan a file system and return the file details"""
import logging
import os
import sys

from lost_cat.utils.tag_anon import TagAnon

logger = logging.getLogger(__name__)

class BaseParser():
    """A base class for a single file parser or converter"""
    def __init__(self, uri : str = None, bytes_io: bytes = None, settings: dict = None) -> None:
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = BaseParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Settings: %s", self.settings)

        # init the tags lists
        self._anonobj = None
        self._alias_tags = {}
        self._groups_tags = {}
        self._metadata_tags = {}

    @property
    def name(self):
        """Returns the class name"""
        return self._name

    @property
    def version(self):
        """Returns the version"""
        return self._version

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "anonimizer": self.set_anonimizer,
            "tags_alias": self.set_alias_tags,
            "tags_groups": self.set_groups_tags,
            "tags_metadata": self.set_metadata_tags,
            "parser": self.parser
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{},
            "uritypes": ["file"],
            "filter":[
                {
                    "table": "URIMD",
                    "key": "ext",
                    "values": []
                }
            ]
        }

    def set_anonimizer(self, anonimizer: TagAnon) -> None:
        """Allows for the metadata tags to be anonymized"""
        self._anonobj = anonimizer

    def set_alias_tags(self, tags: dict) -> None:
        """Allows for the metadata tags to be anonymized
            <tag to rename>: <tag new name>
        '"""
        self._alias_tags = tags

    def set_groups_tags(self, tags: list) -> None:
        """Sets the metadata tags to be used for grouping
        The grouping is used to organize the structure"""
        self._groups_tags = tags

    def set_metadata_tags(self, tags: list):
        """Sets the tags to use for general metadata"""
        self._metadata_tags = tags

    def close(self, force: bool = False, block: bool = False, timeout: int = -1):
        """will close the current file and connectors"""
        pass

    def parser(self) -> dict:
        """will parser the open file and retrn the result"""
        _data = {}
        return _data



"""the base class for the scanners"""
import logging
import multiprocessing as mp
from time import sleep, time

from lost_cat.utils.tag_anon import TagAnon

logger = logging.getLogger(__name__)

class UnableToLoadProcessor(Exception):
    """Exception to handle the class load failed! """

class InvalidURIGiven(Exception):
    """Exception to raise the URL is incorrect """

class BaseProcessor():
    """This will create the core functions
    """
    def __init__(self, settings: dict = None):
        """sets up the class and attaches the queues"""
        self.settings = settings
        self.input = None
        self.output = None
        self.processorid = -1
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"
        self._semiphore = f"DONE: {self._name}"

        # init the tags lists
        self._anonobj = None
        self._groups_tags = []
        self._metadata_tags = []
        self._alias_tags = {}

    @property
    def name(self):
        """the freindly name of the class """
        return self._name

    @property
    def version(self):
        """the version of the class """
        return self._version

    @property
    def semiphore(self):
        """the semiphore to check for end the queue reads """
        return self._semiphore

    def set_anonimizer(self, anonimizer: TagAnon) -> None:
        """Allows for the metadata tags to be anonymized"""
        self._anonobj = anonimizer

    def set_alias_tags(self, tags: dict) -> None:
        """Allows for the metadata tags to be anonymized
            <tag to rename>: <tag new name>
        '"""
        self._alias_tags = tags

    def set_groups_tags(self, tags: list) -> None:
        """Sets the metadata tags to be used for grouping
        The grouping is used to organize the structure"""
        self._groups_tags = tags

    def set_metadata_tags(self, tags: list):
        """Sets the tags to use for general metadata"""
        self._metadata_tags = tags

    def in_queue(self, in_queue: mp.Queue):
        """the queue to use for input """
        self.input = in_queue

    def out_queue(self, out_queue: mp.Queue):
        """the queue to use for output """
        self.output = out_queue

    def metadata(self) -> dict:
        """Returns the metadata about the current root
        device """
        return {"error": "not implemented"}

    def scan(self):
        """starts the scanner operation"""

    def fetch(self, uri: str) -> object:
        """will return a file, stream, or binary"""

    def upload(self, data: object, uri:str = None) -> bool:
        """will upload the data to the provided uri, if not uri
        provided will write into the current location"""
        return False

    def close(self, force: bool = False, block: bool = False, timeout: int = -1):
        """will close the objects """
        logger.debug("%s %s %s", force, block, timeout)
        if self.input:
            # check for force and / or block
            if force:
                self.input = None
            else:
                start_time = time()
                self.input.put(self.semiphore)
                while True:
                    sleep(10)
                    if time() - start_time > timeout:
                        logger.debug("Timeout hit %s %s", timeout, time() - start_time)
                        break

            self.input = None

    def build_path(self, uri: str, hint: str = "n/a") -> dict:
        """will accept a uri and will turn it to a dictionary
        for later use.
        and return a dictionary
            "type": str
            "uri": uri
        """

        src = {
            "uri": uri,
            "type": f"CLASS:{self.__class__.__name__}",
            "domain": "<base class>"
        }

        return src

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            "sources": self.in_queue,
            "uris": self.out_queue,
            #"scan": self.scan,

            #"fetch": self.fetch,
            #"upload": self.upload,

            "anonimizer": self.set_anonimizer,
            "tags_alias": self.set_alias_tags,
            "tags_metadata": self.set_metadata_tags,
            "tags_groups": self.set_groups_tags,
        }

    @staticmethod
    def avail_config() -> dict:
        """returns configuration details about the class
            "options":          dict    specific to the called class
            "uritypes":         array   uri_types to filter sources
            "<functionname>":   array   dict
                                        provide a simplified query for the
                                        generation of the funcname that
                                        is being called

                    "table"             str     the table to search
                                        must relate to a URI table
                    "field"             str     the fields to check
                    "select"            str     the value to match as the field
                    "filter"            list    values to filter the value field
            "uri_metadata":     list    a list of fields to send
                                        in the processor class queue

            "version_metadata": list    a list of fields to send
                                        in the processor class queue
            "threads":
                "count":        int     how many threads to create
                "stop":         bool    force stop on close
                "timeout":      int     timeout in seconds
        """
        return {}



""" Base class for the action processor
Teh action is a means to run a operation against a object
in the queues.
"""

import logging

logger = logging.getLogger(__name__)

class BaseAction():
    """ The base action required th4e following:
        - settings: dict of
                information to load and
                initilize the action
        - criteria: dict
                informaiton to proces the object
                and / or its associated information
                to select the object to the action
        - parser: object (class, function)
        - params: dict
                A dict to build the and run against the
                parser obejct
        - destination: dict
                a set of information to sepcifiy the
                output information
                    - name / rename
                    - destination location
    """
    
    def __init__(self) -> None:
        pass


"""this will scan a file system and return the file details"""
import logging
import multiprocessing as mp
import os
import re
import threading as td
from queue import Empty
from socket import gethostname
from urllib.parse import urljoin, urlparse

from azure.storage.blob import (BlobClient, BlobServiceClient, ContainerClient,
                                __version__)
from lost_cat.processors.base_processor import (BaseProcessor, InvalidURIGiven,
                                                UnableToLoadProcessor)
from lost_cat.utils.path_utils import (SourceDoesNotExist, SourceNotHandled,
                                       build_path, get_file_metadata,
                                       get_filename, scan_files)

logger = logging.getLogger(__name__)

class AzureStorageProcessor(BaseProcessor):
    """This class will perform serveral operations:
            process an acceptable uri
            scan a uri and catalog found items there
            fetch an item from a uri
            upload an item to a uri
    """

    def __init__(self, settings: dict = None):
        """"""
        super().__init__(settings=settings)
        self._version = "0.0.2"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"
        self._semiphore = f"DONE: {self._name}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = AzureStorageProcessor.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Semiphore: %s", self._semiphore)
        logger.debug("Settings: %s", self.settings)

        logger.info("Azure Blob Storage v%s", __version__)
        connect_str = os.environ.get(self.settings.get("connvar"))
        if not connect_str:
            logger.error("Unable to fetch connection string! Missing Connvar and envar")
            UnableToLoadProcessor()

        self._bsclient = BlobServiceClient.from_connection_string(connect_str)
        self._bsurl = self._bsclient.url
        self._bsacctname = self._bsclient.account_name

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            #"build": self.build_path,
            "scanner": self.scan,
            #"fetch": self.fetch,
            "upload": self.upload,
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "connvar": "AZURE_STORAGE_CONNSTTR",
            "options":{
                "splitfolders": True,
                "splitextension": True,
                "stats": True,
            },
            "uritypes": [
                "AzureStorage", "AzureStorage.node", "AzureStorage.blob"
            ],
            "filters":[],
            "threads": {
                "count": 1,
                "stop": True, # tells the system to stop when the in queue is empty
                "timeout": 2
            }
        }

    def metadata(self) -> dict:
        """ Returne the properties of the root object"""
        _data = {
            "uri" :     self._bsurl,
            "account":  self._bsacctname,
            "settings": self.settings,
        }

        return _data

    def scan(self):
        """scan the file system use multprocess to
        scan each item in the queue"""
        use_threads = self.settings.get("threads",{}).get("count",5)

        for t_idx in range(use_threads):
            logger.info("Thread: %s",t_idx)
            scan_q = td.Thread(target=self.scan_fs)
            scan_q.start()
            scan_q.join()

    def scan_fs(self):
        """A file scanner function, quick and dirty"""
        t_settings = self.settings.get("threads",{})

        while self.input:
            try:
                q_item = self.input.get(timeout=t_settings.get("timeout")) if self.input else None
                if q_item == self.semiphore:
                    break

                # if the user wants to kill the queue if there are not entries...
                # requires timeout set, otherwise the queue get blocks...
                if not q_item and self.settings.get("threads",{}).get("stop"):
                    break

                uri = q_item
                uriid = -1

                if isinstance(q_item, dict):
                    uri = q_item.get("uri")
                    uriid = q_item.get("uriid")

                _urlparts = urlparse(uri)

                # run the scan of the URL
                _con_client = self._bsclient.get_container_client(uri)

                # see what is in the container...
                _blob_list = _con_client.list_blobs()
                for _blob in _blob_list:
                    _hostname = "{}::{}".format(self.settings.get("connvar"), _blob.container)
                    _uri = "{}{}/{}".format(self._bsurl, _blob.container, _blob.name)
                    if _blob.size > 0:
                        _, _ext = os.path.splitext(_blob.name)
                        _type = ".blob"
                    else:
                        _ext = ''
                        _type = ".node"

                    _data =  {
                        "domain": self._bsurl,

                        "processorid": self.processorid,
                        "uriid": uriid,
                        "uri_type": f"AzureStorage{_type}",
                        "uri": _uri,
                        "deleted": _blob.deleted_time,
                        "metadata": {
                            "ext": _ext,
                            "created": _blob.creation_time,
                            "name": _blob.name,
                            "container": _blob.container,
                            "host": self._bsurl,
                        },
                        "versions": {
                            "modified": _blob.last_modified,
                            "size": _blob.size,
                            "tags": _blob.tags,
                            "etag":	_blob.etag,
                            "metadata":	_blob.metadata,
                        }
                    }

                    self.output.put(_data)

            except Empty:
                break

    def upload(self, destination:str, source: str = None, data:bytes = None):
        """Will upload the data to the given url
        uri will include the account, container, and path"""
        (scheme, netloc, urlpath, params, query, fragment) = urlparse(destination)
        (bsscheme, bsnetloc, bspath, bsparams, bsquery, bsfragment) = urlparse(self._bsurl)

        if netloc != bsnetloc:
            # wrong acct given
            logger.error("Incorrect Account: Given: %s, expected: %s", bsnetloc, netloc)
            raise InvalidURIGiven()

        _ppaths = urlpath.split("/")
        logger.info("Paths:     %s",_ppaths)
        _container = _ppaths[1]
        logger.info("Container: %s", _container)

        _path = "/".join(_ppaths[2:]) if len(_ppaths) > 2 else None
        logger.info("Path:      %s", _path)

        _bclient = self._bsclient.get_blob_client(container=_container, blob=_path)
        if data:
            _bclient.upload_blob(data)
        elif source and os.path.exists(source):
            with open(source, "rb") as data:
                _bclient.upload_blob(data)



import docx2txt
from docx import Document
import logging
import os
import sys

from lost_cat.parsers.base_parser import BaseParser

logger = logging.getLogger(__name__)

class WordParser(BaseParser):
    """Process a word document"""
    def __init__(self, uri: str = None, bytes_io: bytes = None, settings: dict = None) -> None:
        super().__init__(uri=uri, bytes_io=bytes_io, settings=settings)
        self._version = "0.0.1"
        self._name = f"{self.__class__.__name__.lower()} {self._version}"

        if not settings:
            logger.debug("Loading default settings")
            self.settings = WordParser.avail_config()

        logger.debug("Name: %s", self._name)
        logger.debug("Settings: %s", self.settings)

        # file
        self._uri = None
        self._file = None
        if uri:
            self._uri = uri
            self._file = Document(self._uri)
        elif bytes_io:
            self._bytes_io = bytes_io
            self._file = Document(bytes_io)

    def avail_functions(self) -> dict:
        """Returns a dict prointing to the available functions"""
        return {
            #"anonimizer": self.set_anonimizer,
            #"tags_alias": self.set_alias_tags,
            #"tags_metadata": self.set_metadata_tags,
            #"tags_groups": self.set_groups_tags,
            "parser": self.parser,
            "content": self.get_content,
        }

    @staticmethod
    def avail_config() -> dict:
        """returns default configuration details about the class"""
        return {
            "options":{},
            "uritypes": ["file"],
            "source":[
                {
                    "table": "URIMD",
                    "key": "ext",
                    "values": [".docx"]
                }
            ]
        }

    def close(self, force: bool = False, block: bool = False, timeout: int = -1):
        """will close the """
        if self._file:
            self._file = None

    def parser(self) -> dict:
        """will parser the open file and retrn the result"""
        return self.get_metadata()

    def get_metadata(self):
        """ """
        cp = self._file.core_properties
        return {
            "author": 	        cp.author,
            "category": 	    cp.category,
            "comments": 	    cp.comments,
            "content_status": 	cp.content_status,
            "created": 	        cp.created,
            "identifier": 	    cp.identifier,
            "keywords": 	    cp.keywords,
            "language": 	    cp.language,
            "last_modified_by": cp.last_modified_by,
            "last_printed": 	cp.last_printed,
            "modified": 	    cp.modified,
            "revision": 	    cp.revision,
            "subject": 	        cp.subject,
            "title": 	        cp.title,
            "version": 	        cp.version,
        }

    def get_content(self):
        """ """
        _text = docx2txt.process(self._uri)
        _lines = [x.strip() for x in _text.split("\n") if x.strip()]

        return {
            "type": "lines",
            "lines": _lines
        }



""" Utilities to build and manage TRIEs """

import logging
import string

logger = logging.getLogger(__name__)

class WordNode():
    """ A node to store aword in the system"""
    def __init__(self, word:str, depth:int = 0) -> None:
        self.word = word
        self.counter = 0
        self.depth = depth
        self.is_end = False
        self.children = {}

    def __repr__(self) -> str:
        return f"<WordNode [{self.word}]>"

    def fetchchild(self, word: str) -> object:
        """ Will return the child object that
        matches the word"""
        return self.children.get(word)

class TrieWord():
    """ a word Trie root"""

    def __init__(self) -> None:
        """Initializes and create a root node"""
        self.root = WordNode("")

    def insert(self, phrase: str):
        """Insert a phrase, a collection of word
        and will build out the tree accordingly"""
        node = self.root
        phrase = phrase.replace("_", " ").replace("-", " ").translate(str.maketrans('', '', string.punctuation))
        words = [x for x in phrase.split()]

        for idx, word in enumerate(words):
            if not word:
                continue

            if not node:
                break

            word = word.lower().strip()

            if word in node.children:
                node = node.children.get(word)
            else:
                add_node = WordNode(word=word, depth=idx+1)
                node.children[word] = add_node
                node = add_node

        node.is_end = True
        node.counter += 1

    def export(self, node: WordNode = None, depth: int = 0, maxdepth: int = 10):
        """ Returns an indented json object"""
        if depth >= maxdepth:
            return None

        if not node:
            node = self.root

        _data = {}
        for word, childnode in node.children.items():
            _data[word] = self.export(node=childnode, depth=depth+1)

        return _data

def process_sentence(sentence: str, roottrie: TrieWord) -> dict:
    """ Will process the sentene and provide a dict of
    {
        "sentence": <imput sentence>
        "terms": list of phrases found
        "pos": list of elements in the sentence
    }
    """
    tagged = []
    phrases = []
    curr_phrase = []
    node = roottrie
    sentence = sentence.replace(r" \n", " ").replace("  ", " ").replace("_", " ").replace("-", " ").translate(str.maketrans('', '', string.punctuation))
    logger.debug("Sentence: %s", sentence)

    for idx, word in enumerate([x for x in sentence.split()]):
        tags = {}

        logger.debug("\tW:\t%s", word)
        if cnode := node.fetchchild(word.lower()):
            if len(curr_phrase) == 0:
                logger.debug("\t\t\tS =>")

            curr_phrase.append(word)
            node = cnode

        else:
            # only add the phrase if end wordm reset phrase
            if curr_phrase:
                if node.is_end:
                    logger.debug("\t\t\t<= E")
                    phrases.append(" ".join(curr_phrase))

                    tagged.append({
                        "phrase": (" ".join(curr_phrase)),
                        "start":  idx - len(curr_phrase) + 1,
                        "length": len(curr_phrase),
                    })
                else:
                    if len(curr_phrase) == 1:
                        tagged.append({
                            "word": curr_phrase[0],
                            "start":  idx,
                            "length": len(curr_phrase),
                        })

                curr_phrase = []

            logger.debug("\t\t\t<> C")
            node = roottrie

            # check that the current word
            if cnode := node.fetchchild(word.lower()):
                logger.debug("\t\t\tS =>")
                curr_phrase.append(word)
                node = cnode
            else:
                tags = {
                    "word": word,
                    "start": idx+1,
                    "length": 1,
                }

        if len(curr_phrase) == 0:
            tagged.append(tags)
        #markup = [w for d["word"] in ]

    # catch end...
    if curr_phrase:
        if node.is_end:
            phrases.append(" ".join(curr_phrase))
            tagged.append({
                "phrase": (" ".join(curr_phrase)),
                "start":  idx - len(curr_phrase) + 1 ,
                "length": len(curr_phrase),
            })
        elif len(curr_phrase) == 1:
            tagged.append({
                "word": curr_phrase[0],
                "start":  idx,
                "length": len(curr_phrase),
            })
        else:
            logger.error("Curre phrase %s in sentence %s is mislabeled!", curr_phrase, sentence)

    return {
        "sentence": sentence,
        "tagged": tagged,
        "phrases": phrases,
    }



"""A mnodule for the Phrase Tool code and class"""
import logging
import re

logger = logging.getLogger(__name__)

class PhraseTool:
    """
    ---
    Attributes
    ----------

    Methods
    -------

    """

    # instatitate as a class variable
    _regexes = {
        "tcase": re.compile("[A-Z][a-z]+"),
        "abbrv": re.compile("[A-Z]{2,}"),
        "ucase": re.compile("[A-Z]+"),
        "lcase": re.compile("[a-z]+"),
        "code": re.compile(r"[\u4e00-\u9fff]+"),
        "decimal": re.compile(r"\d+\.\d+"),
        "int": re.compile(r"\d+"),
        "group": re.compile(r"[()\[\]{}']+"),
        "formula": re.compile(r"[,!@#$%&+=?]+"),
        "space": re.compile(r"[\. _\-\t]+")
    }
    name = "names"
    version = "0.0.1"

    def __init__(self, phrase: str) -> None:
        self._phrase = phrase

    def __str__(self):
        return f"{self.name} {self.version} <{self._phrase}>"

    def get_metadata(self, phrase: str = None) -> dict:
        """This will return a profile of the file path and name

        Parameters
        ----------

        Returns
        -------
        None
        """
        if not phrase:
            phrase = self._phrase

        positions = {}
        p = {}
        for k,v in self._regexes.items():
            for m in v.finditer(phrase):
                l = len(m.group())
                s = m.start()
                e = m.end()
                if s not in p:
                    p[s] = {
                        "start": s,
                        "end": e,
                        "type": k,
                        "value": m.group()
                    }

                    # remove the found groups with place holders
                    phrase = "{}{}{}".format(phrase[:s], k[0]*l, phrase[e:])
                    if l == 1:
                        positions[s] = k[0]
                    else:
                        positions[s] = "{}{}".format(k[0],l)

        out_phrase = []
        parts = []
        for k, v in sorted(positions.items()):
            out_phrase.append(v.upper())
            parts.append(p.get(k))
        pro_phrase = "".join(out_phrase)

        return {
            "parts": parts,
            "short": re.sub('\\d','', pro_phrase),
            "profile": pro_phrase,
            "expand": phrase
        }

    def get_parts(self, phrase: str = None) -> dict:
        """Uses the hard code parser to break down the phrase"""
        if not phrase:
            phrase = self._phrase

        _rows = []
        _transtr = []
        _shrtstr = []

        block = []
        state = None
        start = 0
        count = 0

        for idx, c in enumerate(phrase):
            if c.isupper():
                _newstate = "u"
            elif c.isalpha():
                _newstate = "l"
            elif c.isdigit():
                _newstate = "d"
            elif c.isspace():
                _newstate = "s"

            elif c in "(){}[]<>":
                _newstate = "b"
            elif c.isprintable():
                _newstate = "p"
            else:
                _newstate = "?"

            # process and update data...
            if _newstate == state or (_newstate =="l" and state=="t"):
                count += 1
                block.append(c)

            elif state == "u" and _newstate == "l":
                # deteect uppercased phrases...
                if len(block) > 1:
                    _transtr.append(f"{state}{count}")
                    _shrtstr.append(state)
                    _rows.append({
                        "type": state,
                        "start": start,
                        "end": idx-2,
                        "block": "".join(block[:-1])
                    })
                    count = 1
                    block = [block[-1]]

                state = "t"
                count += 1
                block.append(c)

            else:
                if state:
                    _transtr.append(f"{state}{count}")
                    _shrtstr.append(state)
                if block:
                    _rows.append({
                        "type": state,
                        "start": start,
                        "end": idx-1,
                        "block": "".join(block)
                    })

                count = 1
                state = _newstate
                start = idx
                block = [c]

        _transtr.append(f"{state}{count}")
        _shrtstr.append(state)

        # handle final state:
        _rows.append({
            "type": state,
            "start": start,
            "end": idx,
            "block": "".join(block)
        })

        return {
            "profile": "".join(_transtr),
            "short": "".join(_shrtstr).replace("s","").replace("p","").replace("?",""),
            "blocks": _rows
            }
